from igraph import *


def read_data(file):
    """Reads a data file in the following form: Lines starting with # or % are command lines and hence ignored.
    Each line consists of at least two entries separated by blanks. The first and the second entry are the vertex names
    of an edge. The third entry is the edge weight. This is not implemented yet.
    The function returns a (i)graph consisting of the edges of the inputfile."""
    graph = Graph()
    name_2_id = dict()
    edges = []
    graph.vs["name"] = []

    # read graph line-wise
    l = range(2)
    for line in file:
        if line.startswith('%') or line.startswith('#'): continue
        # get the two (or three) entries of a line
        entries = line.split()
        # check if there is a new vertex
        for i in l:
            vertex = entries[i]
            if not vertex in name_2_id:
                n = graph.vcount()
                graph.add_vertex()
                name_2_id[vertex] = n
                graph.vs[n]["name"] = vertex

        # add the new edge
        edges += [(name_2_id[entries[0]], name_2_id[entries[1]])]

    graph.add_edges(edges)
    return graph

def sort_graph(in_graph, increasing=True, depth=2, hierarchy=True):
    graph = Graph()
    graph.vs["name"] = []
    indices = range(in_graph.vcount())
    deg2 = []
    square = len(indices)**2
    
    for i in indices:
        if hierarchy:
            d = square*in_graph.degree(i)
        else:
            d = in_graph.degree(i)
        for j in in_graph.neighbors(i):
            if depth<2: break
            d += in_graph.degree(j)
            for k in in_graph.neighbors(j):
                if depth<3: break
                d += in_graph.degree(k)
        deg2.append(d)
    if increasing:
        sizes = sorted(zip(deg2,indices), reverse=True)
    else:
        sizes = sorted(zip(deg2,indices), reverse=False)

    deg, indices = zip(*sizes)

    edges = []
    old_to_new = dict()
    for i in range(len(deg)):
        old_to_new [indices[i]]=i
        graph.add_vertex()
        graph.vs[i]["name"]= in_graph.vs[indices[i]]["name"]

    for e in in_graph.es:
        u = e.source
        v = e.target
        edges += [(old_to_new[u], old_to_new[v])]

    graph.add_edges(edges)
        
    return graph
