from __future__ import print_function
from igraph import *
import data.read_data_white as read
import src.enumerate_subgraphs as enu
import src.eval_func as eval
import src.twins as twins
import src.greedy as greedy
from time import *
import sys
import argparse


# the first file contains the program parameters (which enumeration algorithm, how often evaluate, ...)
# the second file contains the graph, parameter k and the problem (with parameters)
parser = argparse.ArgumentParser(description='Fixcon: Find optimal subgraph.')
parser.add_argument('algo', type=str, help='Specify enumeration algorithm, options are pivot-iter, simple-iter and kavosh-iter')
parser.add_argument('graph_file', type=str, help='graph file')
parser.add_argument('parameter', type=int, help='subgraph size')
parser.add_argument('problem', type=str, help='problem and problem parameters, seperated by ","')
parser.add_argument('time_limit', type=int, help='time limit in seconds for each instance')
args = parser.parse_args()

# convert input in float if possible
def isfloat(value):
    try:
        float(value)
        return float(value)
    except ValueError:
        return value

# get graph_problem, parameter k and problem
data_file = args.graph_file
k = args.parameter
problem_paras = args.problem.split(",")
problem = int(problem_paras[0])
parameters = map(int, problem_paras[1:len(problem_paras)])

# set problem
eval.set_problem(problem, parameters, k)
# load file and built graph
file = open(data_file.rstrip())
name = str(data_file).split('/')[-1]
graph = read.read_data(file)
graph.simplify()
file.close()

print("Input instance has", graph.vcount(), "vertices and", graph.ecount(), "edges.")
print(clock())

t1 = clock()

monotone = False
limit = True
crit = 0
children = 0

# now calculate greedy lower bounds, if maximum is true so optimum is already reached
# only if time left
if limit is True:
    value, best_graph, maximum = greedy.calc_random_lower_bound(graph, k, t1 + args.time_limit)
    print("Heuristic lower bound",value,"computed in",clock(),"seconds.")
else:
    value = -sys.maxsize + 10 * k
    best_graph = Graph()
    best_graph.vs["name"] = []
    maximum = False

if eval.edge_monotone:
    graph, crit, children, limit = twins.crit_cliques_edge_monotone(graph, k, t1 + args.time_limit)
    monotone = True
else:
    graph, crit = twins.crit_cliques(graph, k)

filename_log = "results/"+args.algo+"/"+name+"."+str(k)+"."+str(problem)+".log"
log_file = open(filename_log, 'w+')
# write results in result file
out_line = (str(problem).rjust(11) + ', '.join(map(str, parameters)).rjust(14) + name.rjust(40) + str(graph.vcount()).rjust(9) + str(graph.ecount()).rjust(11) + str(k).rjust(3))

# experiments
# distinguish whether we use color coding or not
if maximum == True:
    r1 = value
    nodes = 0
    resgraph = best_graph
else:
    # choose evaluation algorithm from dictionary of parameters
    if monotone == True:
        if limit == True:
            r1, resgraph, nodes = enu.enumeration_function[args.algo+"_m"](graph, k, t1 + args.time_limit, crit, children, value)
            if r1 == value:
                resgraph = best_graph
        else:
            r1 = value
            resgraph = best_graph
            nodes = " Calculation of Dom Dag took to much time "
    else:
        r1, resgraph, nodes = enu.enumeration_function[args.algo](graph, k, t1 + args.time_limit, crit, value)
        if r1 == value:
            resgraph = best_graph
t2 = clock()
out_line += (str(round(t2 - t1, 3)).rjust(15) + str(r1).rjust(21))+str(nodes).rjust(20)+" "
out_line += ', '.join(map(str, resgraph.vs["name"]))+"\n"
log_file.write(out_line)
log_file.close()

