#!/usr/bin/env python

from __future__ import print_function
from gurobipy import *

import sys
import time
import networkx as nx

def mycallback(model, where):
    if where != GRB.Callback.MIPSOL:
        pass
    else:
        #print(model.cbGet(GRB.Callback.MIPSOL_OBJ))
        sol = model.cbGetSolution(model._nodevars)
        sol_verts = set()
        for i in range(len(model._nodevars)):
            if(sol[i]>0.1):
                sol_verts.add(model._nodevars[i].varName)

        #print(sol_verts)
        G = model._G
        GS = nx.subgraph(G,sol_verts)
        #print("Edge number: " +  str(GS.number_of_edges()))
        if not nx.is_connected(GS):
            #print("Not connected")
            cc = nx.connected_components(GS)
            for c in cc:
                expr = LinExpr()
                Nc = set()
                for v in c:
                    Nc.update(G.neighbors(v))
                Nc.difference_update(c)
                for v in c:
                    expr = LinExpr()
                    for u in Nc:
                        expr.add(model.getVarByName(str(u)))
                    model.cbLazy(expr,GRB.GREATER_EQUAL,m.getVarByName(str(v)))
                    
        
def get_str(a,b):
    if a<b:
        return str(a)+","+str(b)
    else: 
        return str(b)+","+str(a)

def add_summand(expr,var):
    if (var.x>0.0):
        expr.add(1-var)
    else: 
        expr.add(var)

sys.stdout.write("#")
        
G = nx.Graph()
k = float(sys.argv[1])
for line in open(sys.argv[2]):
    #print(line)
    fields=line.split()
    if not (line.startswith('#') or line.startswith('%')):
        u = fields[0]
        v = fields[1]
        G.add_node(u)
        G.add_node(v)
        if u != v:
            G.add_edge(u,v)
        
#print(G.number_of_nodes(),file=sys.stderr)
#print(G.number_of_edges(),file=sys.stderr)

start = time.time()

try:

    # Create a new model
    m = Model("Densest k Subgraph")
    m.Params.OutputFlag = 0
    m.Params.LazyConstraints = 1
    m.setParam('TimeLimit', 600)
    # Create variables
    obj = LinExpr();
    
    for u in G.nodes():
        var = m.addVar(vtype=GRB.BINARY, name=str(u))

    m.update()
    m._nodevars = m.getVars()
    m._G = G
    
    for (u,v) in G.edges():
        var_e = m.addVar(vtype=GRB.BINARY, name=get_str(u,v))
        obj += var_e;
    # Integrate new variables
    m.update()

    # Set objective
    m.setObjective(obj, GRB.MAXIMIZE)
    
    # Add constraints: each edge can be selected only if endpoints are selected 
    for (u,v) in G.edges():
        xu = m.getVarByName(str(u))
        xv = m.getVarByName(str(v))
        xuv = m.getVarByName(get_str(u,v))
        m.addConstr(xuv <= xu)
        m.addConstr(xuv <= xv)
    # Add constraints: at most k vertices are selected 
    #print("Adding size constraint")
    expr = LinExpr()        
    for u in G.nodes():
        var = m.getVarByName(str(u))
        expr += var
    m.addConstr(expr, GRB.EQUAL, k, 'k-card')

    m.update()
    
    m.setParam("BestObjStop", k*(k-1)/2.0)

    m.optimize(mycallback)

    #print(m.ObjVal)
   
    # solved=False
    
    vertex_set = []

    #default infeasible / no solution found
    obj = -1
    sol_string = ""
    
    if m.Status!=3 and m.Status!=4 and m.Status != 9:
        for v in m.getVars():
            if (len(v.varName.split(","))<2 and v.x>0.99):
                vertex_set.append(v.varName)
        sol_string=','.join(vertex_set)
    logfile = open(sys.argv[3]+"results.log",'a')
    logfile.write('\t'.join(['1',"",(sys.argv[2].split('/'))[-1],str(G.number_of_nodes()),str(G.number_of_edges()),str(int(k)),"{:10.2f}".format(time.time()-start),str(m.ObjVal)," ", sol_string+"\n"]))
    logfile.close()
             
except GurobiError as e:
    print('Gurobi error ' + str(e.errno) + ": " + str(e.message))
