from __future__ import print_function
from igraph import *
from time import *
import src.eval_func as ev
import src.twins as twins
from itertools import combinations

# input is a list or a set of vertex IDs, output is a list of vertex names
def get_names(graph, vertex_ids):
    lis = []
    for id in vertex_ids:
        lis.append(graph.vs[id]["name"])
    return lis

def update_forbidden(i,crit,graph,forb_vertices):
    for vertex in crit[graph.vs[i]["crit"]]:
        forb_vertices[vertex] = True
    

def update_forbidden_monotone(i,crit,children,graph,forb_vertices):
    for vertex in crit[graph.vs[i]["crit"]]:
        forb_vertices[vertex] = True
    for vertex in children[i]:
        forb_vertices[vertex] = True
        
def update_graph(graph,i,del_edges):
    for j in graph.neighbors(i):
        del_edges.append((i,j))
    if (i % math.floor(math.sqrt(graph.vcount())) == 0):
        graph.delete_edges(del_edges)
        del_edges = []
    return del_edges

        
def print_progress(i,n,v):
    step = 2
    if n>200: step = int(n/100)
    if (i % step == 0):
        print("{0:.2f}".format(float(n-i)/float(n)*100.0)+"% of vertices processed in",clock(),"seconds. Current lower bound:",v)
        

def pivot_iter_noreturn_m(graph, k, time_max, crit, children, value):
    del_edges = []
    best_graph = Graph()
    best_graph.vs["name"] = []
    current_graph = Graph()
    current_graph.add_vertex()
    nodes = 0
    # colors for vertices (igraph is too slow)
    forb_vertices = graph.vcount() * [False]
    colors = graph.vcount()*[-1]
    # consider only graphs with at least k vertices
    for i in range(graph.vcount() - 1, k - 2, -1):
        print_progress(i,graph.vcount(),value)
        if forb_vertices[i] is True:
            continue
        current_graph.vs[0]["id"] = i
        # check if this component has size at least k
        nodes += 1
        # subgraph_set
        subgraph_set = [i]
        # subgraph set as set
        sub_set = {i: 0}
        # extension set
        extension = []
        # list of lists of all exclusive neighbors
        ex_neighs = [[]]
        # color closed neighborhood of start vertex
        colors[i] = 0
        forb_vertices[i] = True
        for vertex in graph.neighbors(i):
            if vertex > i:
                break
            colors[vertex] = 1
            extension.append(vertex)
            ex_neighs[0].append(vertex)
        # lists for the pointers of the extension set
        pointers = [0]
        # length of the extension set according to each added vertex
        ext_lengths = [len(ex_neighs[0])]
        # array to save in which extension set the pointer is
        still_neighbors = [[0]]
        # subgraph size
        sub_size = 1
        # array to use forbidden vertices from critical cliques, we use color "-2" to test this efficiently
        # in forbidden only save the representatives and while retoring check for "-2"
        forbidden = [[]]
        # enumerate all subgraphs containing vertex i
        while subgraph_set != []:
            # for each vertex in the extension set, create a new branch
            # if the actual pointer points to null (-1), then the corresponding extension set is empty
            while pointers[-1] < len(extension):
                # check time limit
                if clock() > time_max:
                    return value, best_graph, nodes
                first = pointers[-1]
                vertex = extension[first]
                # move pointer one to the left
                pointers[-1] += 1
                # check if some vertices do not have some exclusive neighbors now
                while still_neighbors[-1] and pointers[-1] > ext_lengths[still_neighbors[-1][0]]:
                    still_neighbors[-1].pop(0)
                # do nothing if this vertex is forbidden
                if forb_vertices[vertex] is True:
                    continue
                # create next child
                nodes += 1
                forb_vertices[vertex] = True
                forbidden[-1].append(vertex)
                sub_size_minus = sub_size
                sub_size += 1
                ex_neighs.append([])
                edges = []
                current_graph.add_vertex()
                current_graph.vs[sub_size_minus]["id"] = vertex
                found = False
                for neig in graph.neighbors(vertex):
                    if neig > i:
                        break
                    if forb_vertices[neig] is False:
                        found = True
                        if colors[neig] == -1:
                            colors[neig] = sub_size
                            ex_neighs[-1].append(neig)
                    if neig in sub_set:
                        edges.append([sub_set[neig], sub_size_minus])
                current_graph.add_edges(edges)
                still_neighbors.append(still_neighbors[-1][:])
                if found == True:
                    still_neighbors[-1].append(sub_size_minus)
                # check if adding this vertex leads to a solution
                if sub_size == k:
                    new_value = ev.evaluate(current_graph,graph,forb_vertices,still_neighbors[-1])
                    if new_value > value:
                        value = new_value
                        best_graph = current_graph.copy()
                        subgraph_set.append(vertex)
                        best_graph.vs["name"] = get_names(graph, subgraph_set)
                        subgraph_set.pop()
                        # check if user_ub is matched
                        if ev.check_user_ub(value):
                            return value, best_graph, nodes
                    sub_size -= 1
                    # forbid each vertex in the same critical clique
                    for member in crit[graph.vs[vertex]["crit"]]:
                        if member >= i:
                            break
                        if forb_vertices[member] is False:
                            forbidden[-1].append(member)
                            forb_vertices[member] = True
                    # also forbid each vertex whose neighborhood is a subset of this vertex
                    for member in children[vertex]:
                        if forb_vertices[member] is False:
                            forbidden[-1].append(member)
                            forb_vertices[member] = True
                    for neig in ex_neighs[-1]:
                        colors[neig] = -1
                    ex_neighs.pop()
                    current_graph.delete_vertices(sub_size)
                    still_neighbors.pop()
                # otherwise check if adding this vertex can increase the value, if this is not possible, return to its parent
                elif ev.eval_add_vertices(current_graph, graph, k - sub_size, still_neighbors[-1], ext_lengths[-1] - first, value, forb_vertices):
                    sub_size -= 1
                    # forbid each vertex in the same critical clique
                    for member in crit[graph.vs[vertex]["crit"]]:
                        if member >= i:
                            break
                        if forb_vertices[member] is False:
                            forbidden[-1].append(member)
                            forb_vertices[member] = True
                    # also forbid each vertex whose neighborhood is a subset of this vertex
                    for member in children[vertex]:
                        if forb_vertices[member] is False:
                            forbidden[-1].append(member)
                            forb_vertices[member] = True
                    for neig in ex_neighs[-1]:
                        colors[neig] = -1
                    ex_neighs.pop()
                    current_graph.delete_vertices(sub_size)
                    still_neighbors.pop()
                # otherwise create new enumeration tree node
                else:
                    sub_set.update({vertex: sub_size_minus})
                    subgraph_set.append(vertex)
                    for member in ex_neighs[-1]:
                        extension.append(member)
                    # new pointer points to old pointer
                    pointers.append(pointers[-1])
                    forbidden.append([])
                    ext_lengths.append(ext_lengths[-1] + len(ex_neighs[-1]))
            # now the last pointer points to null, so restore the parent
            pointers.pop()
            # make all forbidden critical cliques accessible again
            index = 0
            for member in forbidden[-1]:
                forb_vertices[member] = False
                index += 1
            forbidden.pop()
            for vertex in ex_neighs[-1]:
                # also remove this vertex from the extension set
                extension.pop()
                colors[vertex] = -1
            ex_neighs.pop()
            sub_size -= 1
            vertex = subgraph_set.pop()
            sub_set.pop(vertex)
            ext_lengths.pop()
            still_neighbors.pop()
            if subgraph_set != []:
                current_graph.delete_vertices(sub_size)
                for member in crit[graph.vs[vertex]["crit"]]:
                    if member >= i:
                        break
                    if forb_vertices[member] is False:
                        forbidden[-1].append(member)
                        forb_vertices[member] = True
                # also forbid each vertex whose neighborhood is a subset of this vertex
                for member in children[vertex]:
                    if forb_vertices[member] is False:
                        forbidden[-1].append(member)
                        forb_vertices[member] = True
        # now the root i and each vertex in that critical clique are forbidden
        update_forbidden_monotone(i,crit,children,graph,forb_vertices)
        twins.update_twin_relation_monotone(graph, crit, children, i, forb_vertices)
        del_edges = update_graph(graph,i,del_edges)
    return value, best_graph, nodes

# new version of pivot with extension set in "other direction"
def pivot_iter_noreturn(graph, k, time_max, crit, value):
    del_edges = []
    best_graph = Graph()
    best_graph.vs["name"] = []
    current_graph = Graph()
    current_graph.add_vertex()
    nodes = 0
    # colors for vertices (igraph is too slow)
    forb_vertices = graph.vcount() * [False]
    colors = graph.vcount()*[-1]
    # consider only graphs with at least k vertices
    for i in range(graph.vcount() - 1, k - 2, -1):
        print_progress(i,graph.vcount(),value)
        if forb_vertices[i] is True:
            continue
        current_graph.vs[0]["id"]=i
        # check if this component has size at least k
        nodes += 1
        # subgraph_set
        subgraph_set = [i]
        # subgraph set as set
        sub_set = {i: 0}
        # extension set
        extension = []
        # list of lists of all exclusive neighbors
        ex_neighs = [[]]
        # color closed neighborhood of start vertex
        colors[i] = 0
        forb_vertices[i] = True
        for vertex in graph.neighbors(i):
            if vertex > i:
                break
            colors[vertex] = 1
            extension.append(vertex)
            ex_neighs[0].append(vertex)
        # lists for the pointers of the extension set
        pointers = [0]
        # length of the extension set according to each added vertex
        ext_lengths = [len(ex_neighs[0])]
        # array to save in which extension set the pointer is
        still_neighbors = [[0]]
        # subgraph size
        sub_size = 1
        # array to use forbidden vertices from critical cliques, we use color "-2" to test this efficiently
        # in forbidden only save the representives and while retoring check for "-2"
        forbidden = [[]]
        # enumerate all subgraphs containing vertex i
        while subgraph_set != []:
            # for each vertex in the extension set, create a new branch
            # if the actual pointer points to null (-1), then the corresponding extension set is empty
            while pointers[-1] < len(extension):
                # check time limit
                if clock() > time_max:
                    return value, best_graph, nodes
                first = pointers[-1]
                vertex = extension[first]
                # move pointer one to the left
                pointers[-1] += 1
                # check if some vertices do not have some exclusive neighbors now
                while still_neighbors[-1] and pointers[-1] > ext_lengths[still_neighbors[-1][0]]:
                    still_neighbors[-1].pop(0)
                # do nothing if this vertex is forbidden
                if forb_vertices[vertex] is True:
                    continue
                # create next child
                nodes += 1
                forb_vertices[vertex] = True
                forbidden[-1].append(vertex)
                sub_size_minus = sub_size
                sub_size += 1
                ex_neighs.append([])
                current_graph.add_vertex()
                current_graph.vs[sub_size_minus]["id"] = vertex
                edges = []
                found = False
                for neig in graph.neighbors(vertex):
                    if neig > i:
                        break
                    if forb_vertices[neig] is False:
                        found = True
                        if colors[neig] == -1:
                            colors[neig] = sub_size
                            ex_neighs[-1].append(neig)
                    if neig in sub_set:
                        edges.append([sub_set[neig], sub_size_minus])
                current_graph.add_edges(edges)
                still_neighbors.append(still_neighbors[-1][:])
                if found == True:
                    still_neighbors[-1].append(sub_size_minus)
                # check if adding this vertex leads to a solution
                if sub_size == k:
                    new_value = ev.evaluate(current_graph,graph,forb_vertices,still_neighbors[-1])
                    if new_value > value:
                        value = new_value
                        best_graph = current_graph.copy()
                        subgraph_set.append(vertex)
                        best_graph.vs["name"] = get_names(graph, subgraph_set)
                        subgraph_set.pop()
                        # check if user_ub is matched
                        if ev.check_user_ub(value):
                            return value, best_graph, nodes
                    sub_size -= 1
                    # forbid each vertex in the same critical clique
                    for member in crit[graph.vs[vertex]["crit"]]:
                        if member >= i:
                            break
                        if forb_vertices[member] is False:
                            forbidden[-1].append(member)
                            forb_vertices[member] = True
                    for neig in ex_neighs[-1]:
                        colors[neig] = -1
                    ex_neighs.pop()
                    current_graph.delete_vertices(sub_size)
                    still_neighbors.pop()
                # otherwise check if adding this vertex can increase the value, if this is not possible, return to its parent
                elif ev.eval_add_vertices(current_graph, graph, k - sub_size, still_neighbors[-1], ext_lengths[-1] - first, value,forb_vertices):
                    sub_size -= 1
                    # forbid each vertex in the same critical clique
                    for member in crit[graph.vs[vertex]["crit"]]:
                        if member >= i:
                            break
                        if forb_vertices[member] is False:
                            forbidden[-1].append(member)
                            forb_vertices[member] = True
                    for neig in ex_neighs[-1]:
                        colors[neig] = -1
                    ex_neighs.pop()
                    current_graph.delete_vertices(sub_size)
                    still_neighbors.pop()
                # otherwise create new enumeration tree node
                else:
                    sub_set.update({vertex: sub_size_minus})
                    subgraph_set.append(vertex)
                    for member in ex_neighs[-1]:
                        extension.append(member)
                    # new pointer points to old pointer
                    pointers.append(pointers[-1])
                    forbidden.append([])
                    ext_lengths.append(ext_lengths[-1] + len(ex_neighs[-1]))
            # now the last pointer points to null, so restore the parent
            pointers.pop()
            # make all forbidden critical cliques accessible again
            index = 0
            for member in forbidden[-1]:
                forb_vertices[member] = False
                index += 1
            forbidden.pop()
            for vertex in ex_neighs[-1]:
                # also remove this vertex from the extension set
                extension.pop()
                colors[vertex] = -1
            ex_neighs.pop()
            sub_size -= 1
            vertex = subgraph_set.pop()
            sub_set.pop(vertex)
            ext_lengths.pop()
            still_neighbors.pop()
            if subgraph_set != []:
                current_graph.delete_vertices(sub_size)
                for member in crit[graph.vs[vertex]["crit"]]:
                    if member >= i:
                        break
                    if forb_vertices[member] is False:
                        forbidden[-1].append(member)
                        forb_vertices[member] = True
        # now the root i and each vertex in that critical clique are forbidden
        update_forbidden(i,crit,graph,forb_vertices)
        twins.update_twin_relation(graph,crit,i,forb_vertices)
        del_edges = update_graph(graph,i,del_edges)
    return value, best_graph, nodes




# new version of simple without recursion and without the pruning rule
def simple_iter_noreturn(graph, k, time_max, crit, value):
    del_edges = []
    best_graph = Graph()
    best_graph.vs["name"] = []
    current_graph = Graph()
    current_graph.add_vertex()
    nodes = 0
    # colors for vertices (igraph is too slow)
    forb_vertices = graph.vcount() * [False]
    colors = graph.vcount()*[-1]
    # consider only graphs with at least k vertices
    for i in range(graph.vcount() -1, k - 2, -1):
        print_progress(i,graph.vcount(),value)
        if forb_vertices[i] is True:
            continue
        # check if this component has size at least k
        nodes += 1
        # subgraph_set
        subgraph_set = [i]
        # subgraph set as set
        sub_set = {i: 0}
        # extension set
        extension = []
        # list of lists of all exclusive neighbors
        ex_neighs = [[]]
        # color closed neighborhood of start vertex
        colors[i] = 0
        forb_vertices[i] = True
        for vertex in graph.neighbors(i):
            if vertex >= i:
                break
            colors[vertex] = 1
            extension.append(vertex)
            ex_neighs[0].append(vertex)
        # lists for the pointers of the extension set
        pointers = [extension.__len__()-1]
        # we save the index in whose closed neighborhood the actual pointers are, to perform the jumps
        poi_col = [1]
        # vertices which have still neighbors
        still_neighbors = [[0]]
        # subgraph size
        sub_size = 1
        # array to use forbidden vertices from critical cliques, we use color "-2" to test this efficiently
        # in forbidden only save the representatives and while restoring check for "-2"
        forbidden = [[]]
        # enumerate all subgraphs containing vertex i
        while subgraph_set != []:
            # for each vertex in the extension set, create a new branch
            # if the actual pointer points to null (-1), then the corresponding extension set is empty
            while pointers[-1] > -1:
                # check time limit
                if clock() > time_max:
                    return value, best_graph, nodes
                last = pointers[-1]
                vertex = extension[last]
                # move pointer one to the left
                pointers[-1] -= 1
                act_vertex = pointers[-1]
                # check if vertex is an exclusive neighbor of the vertex last added to the subgraph set
                if colors[vertex] == sub_size:
                    # if yes delete this vertex from the extension set
                    extension.pop()
                # check if pointer jumps, in other words check if the new next vertex in the extension set has a different color than the current vertex
                if act_vertex > -1:
                    if colors[extension[act_vertex]] < poi_col[-1]:
                        # jump with pointer (we stay at the same point if the color of the actual vertex is one less)
                        pointers[-1] = pointers[poi_col[-1]-2]
                        # update color
                        poi_col[-1] = colors[extension[pointers[-1]]]
                # do nothing if this vertex is forbidden
                if forb_vertices[vertex] is True:
                    continue
                # create next child
                nodes += 1
                forb_vertices[vertex] = True
                forbidden[-1].append(vertex)
                sub_size_minus = sub_size
                sub_size += 1
                ex_neighs.append([])
                found1 = False
                found2 = False
                edges = []
                current_graph.add_vertex()
                for neig in graph.neighbors(vertex):
                    if neig > i:
                        break
                    if forb_vertices[neig] is False:
                        found2 = True
                        if colors[neig] == -1:
                            colors[neig] = sub_size
                            ex_neighs[-1].append(neig)
                            found1 = True
                    if neig in sub_set:
                        edges.append([sub_set[neig], sub_size_minus])
                current_graph.add_edges(edges)
                still_neighbors.append(still_neighbors[-1][:])
                if found2 == True:
                    still_neighbors[-1].append(sub_size_minus)
                # check if adding this vertex leads to a solution
                if sub_size == k:
                    new_value = ev.evaluate(current_graph)
                    if new_value > value:
                        value = new_value
                        best_graph = current_graph.copy()
                        subgraph_set.append(vertex)
                        best_graph.vs["name"] = get_names(graph, subgraph_set)
                        subgraph_set.pop()
                        # check if user_ub is matched
                        if ev.check_user_ub(value):
                            return value, best_graph, nodes
                    sub_size -= 1
                    # forbid each vertex in the same critical clique
                    for member in crit[graph.vs[vertex]["crit"]]:
                        if member >= i:
                            break
                        if forb_vertices[member] is False:
                            forbidden[-1].append(member)
                            forb_vertices[member] = True
                    for neig in ex_neighs[-1]:
                        colors[neig] = -1
                    ex_neighs.pop()
                    current_graph.delete_vertices(sub_size)
                    still_neighbors.pop()
                # otherwise check if adding this vertex can increase the value, if this is not possible, return to its parent
                elif ev.eval_add_vertices(current_graph, k - sub_size, still_neighbors[-1], last, value):
                    sub_size -= 1
                    # forbid each vertex in the same critical clique
                    for member in crit[graph.vs[vertex]["crit"]]:
                        if member >= i:
                            break
                        if forb_vertices[member] is False:
                            forbidden[-1].append(member)
                            forb_vertices[member] = True
                    for neig in ex_neighs[-1]:
                        colors[neig] = -1
                    ex_neighs.pop()
                    current_graph.delete_vertices(sub_size)
                    still_neighbors.pop()
                # otherwise create new enumeration tree node
                else:
                    subgraph_set.append(vertex)
                    sub_set.update({vertex: sub_size_minus})
                    for member in ex_neighs[-1]:
                        extension.append(member)
                    # if there is an exclusive neighbor, the new pointer points to him, otherwise to the old pointer
                    if found1 == True:
                        pointers.append(extension.__len__()-1)
                        poi_col.append(sub_size)
                    else:
                        pointers.append(pointers[-1])
                        poi_col.append(poi_col[-1])
                    forbidden.append([])
            # now the last pointer points to null, so restore the parent
            pointers.pop()
            poi_col.pop()
            # make all forbidden critical cliques accessible again
            index = 0
            for member in forbidden[-1]:
                forb_vertices[member] = False
                index += 1
            forbidden.pop()
            for vertex in ex_neighs[-1]:
                colors[vertex] = -1
            ex_neighs.pop()
            sub_size -= 1
            still_neighbors.pop()
            vertex = subgraph_set.pop()
            sub_set.pop(vertex)
            if subgraph_set != []:
                current_graph.delete_vertices(sub_size)
                for member in crit[graph.vs[vertex]["crit"]]:
                    if member >= i:
                        break
                    if forb_vertices[member] is False:
                        forbidden[-1].append(member)
                        forb_vertices[member] = True
        # now the root i and each vertex in that critical clique are forbidden
        update_forbidden(i,crit,graph,forb_vertices)
        twins.update_twin_relation(graph,crit,i,forb_vertices)
        del_edges = update_graph(graph, i, del_edges)
    return value, best_graph, nodes



# new version of simple without recursion and without the pruning rule for edge monotone properties to use neighborhood relations
def simple_iter_noreturn_m(graph, k, time_max, crit, children, value):
    del_edges = []
    best_graph = Graph()
    best_graph.vs["name"] = []
    current_graph = Graph()
    current_graph.add_vertex()
    nodes = 0
    # colors for vertices (igraph is too slow)
    colors = graph.vcount()*[-1]
    forb_vertices = graph.vcount()*[False]
    # consider only graphs with at least k vertices
    for i in range(graph.vcount() -1, k - 2, -1):
        print_progress(i,graph.vcount(),value)
        # do nothing if this vertex is forbidden
        if forb_vertices[i] is True:
            continue
        # check if this component has size at least k
        nodes += 1
        # subgraph_set
        subgraph_set = [i]
        # subgraph set as set
        sub_set = {i: 0}
        # extension set
        extension = []
        # list of lists of all exclusive neighbors
        ex_neighs = [[]]
        # color closed neighborhood of start vertex
        colors[i] = 0
        forb_vertices[i] = True
        for vertex in graph.neighbors(i):
            if vertex >= i:
                break
            colors[vertex] = 1
            extension.append(vertex)
            ex_neighs[0].append(vertex)
        # lists for the pointers of the extension set
        pointers = [extension.__len__()-1]
        # we save the index in whose closed neighborhood the actual pointers are, to perform the jumps
        poi_col = [1]
        # vertices which have still neighbors
        still_neighbors = [[0]]
        # subgraph size
        sub_size = 1
        # array to use forbidden vertices from critical cliques
        # we save the forbidden vertices in a separate array
        # in forbidden only save the representatives and while returning check for "-2"
        forbidden = [[]]
        # enumerate all subgraphs containing vertex i
        while subgraph_set != []:
            # for each vertex in the extension set, create a new branch
            # if the actual pointer points to null (-1), then the corresponding extension set is empty
            while pointers[-1] > -1:
                # check time limit
                if clock() > time_max:
                    return value, best_graph, nodes
                last = pointers[-1]
                vertex = extension[last]
                # move pointer one to the left
                pointers[-1] -= 1
                act_vertex = pointers[-1]
                # check if vertex is an exclusive neighbor of the vertex last added to the subgraph set
                if colors[vertex] == sub_size:
                    # if yes delete this vertex from the extension set
                    extension.pop()
                # check if pointer jumps, in other words check if the new next vertex in the extension set has a different color than the current vertex
                if act_vertex > -1:
                    if colors[extension[act_vertex]] < poi_col[-1]:
                        # jump with pointer (we stay at the same point if the color of the actual vertex is one less)
                        pointers[-1] = pointers[poi_col[-1]-2]
                        # update color
                        poi_col[-1] = colors[extension[pointers[-1]]]
                # do nothing if this vertex is forbidden
                if forb_vertices[vertex] is True:
                    continue
                # create next child
                nodes += 1
                forb_vertices[vertex] = True
                forbidden[-1].append(vertex)
                # find the exclusive neighbors of vertex
                ex_neighs.append([])
                found1 = False
                found2 = False
                sub_size_minus = sub_size
                sub_size += 1
                edges = []
                current_graph.add_vertex()
                for neig in graph.neighbors(vertex):
                    if neig > i:
                        break
                    if forb_vertices[neig] is False:
                        found2 = True
                        if colors[neig] == -1:
                            colors[neig] = sub_size
                            ex_neighs[-1].append(neig)
                            found1 = True
                    if neig in sub_set:
                        edges.append([sub_set[neig], sub_size_minus])
                current_graph.add_edges(edges)
                still_neighbors.append(still_neighbors[-1][:])
                if found2 == True:
                    still_neighbors[-1].append(sub_size_minus)
                # check if adding this vertex leads to a solution
                if sub_size == k:
                    sub_size -= 1
                    new_value = ev.evaluate(current_graph)
                    if new_value > value:
                        value = new_value
                        best_graph = current_graph.copy()
                        subgraph_set.append(vertex)
                        best_graph.vs["name"] = get_names(graph, subgraph_set)
                        subgraph_set.pop()
                        # check if user_ub is matched
                        if ev.check_user_ub(value):
                            return value, best_graph, nodes
                    # forbid each vertex in the same critical clique
                    for member in crit[graph.vs[vertex]["crit"]]:
                        if member >= i:
                            break
                        if forb_vertices[member] is False:
                            forbidden[-1].append(member)
                            forb_vertices[member] = True
                    # also forbid each vertex whose neighborhood is a subset of this vertex
                    for member in children[vertex]:
                        if forb_vertices[member] is False:
                            forbidden[-1].append(member)
                            forb_vertices[member] = True
                    for neig in ex_neighs[-1]:
                        colors[neig] = -1
                    ex_neighs.pop()
                    current_graph.delete_vertices(sub_size)
                    still_neighbors.pop()
                # otherwise check if adding this vertex can increase the value, if this is not possible, return to its parent
                elif ev.eval_add_vertices(current_graph, k - sub_size, still_neighbors[-1], last, value):
                    sub_size -= 1
                    # forbid each vertex in the same critical clique
                    for member in crit[graph.vs[vertex]["crit"]]:
                        if member >= i:
                            break
                        if forb_vertices[member] is False:
                            forbidden[-1].append(member)
                            forb_vertices[member] = True
                    # also forbid each vertex whose neighborhood is a subset of this vertex
                    for member in children[vertex]:
                        if forb_vertices[member] is False:
                            forbidden[-1].append(member)
                            forb_vertices[member] = True
                    for neig in ex_neighs[-1]:
                        colors[neig] = -1
                    ex_neighs.pop()
                    current_graph.delete_vertices(sub_size)
                    still_neighbors.pop()
                # otherwise create new enumeration tree node
                else:
                    subgraph_set.append(vertex)
                    sub_set.update({vertex: sub_size_minus})
                    # add new exclusive neighbors to extension set
                    for member in ex_neighs[-1]:
                        extension.append(member)
                    # if there is an exclusive neighbor, the new pointer points to him, otherwise to the old pointer
                    if found1 == True:
                        pointers.append(extension.__len__()-1)
                        poi_col.append(sub_size)
                    else:
                        pointers.append(pointers[-1])
                        poi_col.append(poi_col[-1])
                    forbidden.append([])
            # now the last pointer points to null, so restore the parent
            pointers.pop()
            poi_col.pop()
            # make all forbidden critical cliques (and their children) accessible again
            index = 0
            for member in forbidden[-1]:
                forb_vertices[member] = False
                index += 1
            forbidden.pop()
            for vertex in ex_neighs[-1]:
                colors[vertex] = -1
            vertex = subgraph_set.pop()
            sub_set.pop(vertex)
            still_neighbors.pop()
            sub_size -= 1
            if subgraph_set != []:
                current_graph.delete_vertices(sub_size)
                for member in crit[graph.vs[vertex]["crit"]]:
                    if member >= i:
                        break
                    # it is not a problem that we pop sub_set before; in any case vertex is in N(S), where S is the current subgraph and will not be added again
                    if forb_vertices[member] is False:
                        forbidden[-1].append(member)
                        forb_vertices[member] = True
                # also forbid each vertex whose neighborhood is a subset of this vertex
                for member in children[vertex]:
                    if forb_vertices[member] is False:
                        forbidden[-1].append(member)
                        forb_vertices[member] = True
            ex_neighs.pop()
        # remove color from the root and from all vertices in that critical clique and corresponding children
        update_forbidden_monotone(i,crit,children,graph,forb_vertices)
        twins.update_twin_relation_monotone(graph, crit, children, i,forb_vertices)
        del_edges = update_graph(graph, i, del_edges)
    return value, best_graph, nodes

# new version of kavosh without the break condition and iterative
def kavosh_iter_noreturn_m(graph, k, time_max, crit, children, value):
    # for each fixed vertex v find all connected subgraphs which include v
    # after the enumeration delete vertex v
    del_edges = []
    current_graph = Graph()
    best_graph = Graph()
    best_graph.vs["name"] = []
    nodes = 0
    k_minus = k - 1
    colors = graph.vcount() * [False]
    # consider only graphs with at least k vertices
    for i in range(graph.vcount()-1, k-2, -1):
        print_progress(i,graph.vcount(),value)
        if colors[i] == True:
            continue
        # save the subgraph set
        subgraph_set = [i]
        # subgraph set as set
        sub_set = {i: 0}
        sub_size_minus = 0
        current_graph.add_vertex()
        # set of pointers to save which enumeration tree node corresponds to which subgraph set
        pointers = [0]
        # neighbors sets
        neighs = []
        for neig in graph.neighbors(i):
            if neig > i:
                break
            neighs.append(neig)
        neighbors_sets = [neighs]
        # list of length of number of vertices which will be added if the next iterator is created
        iterator_length = [min(k-1, neighbors_sets[0].__len__())]
        # list of iterators
        iterators = [combinations(neighbors_sets[0], iterator_length[0])]
        # enumerate subgraphs while subgraph set is not empty
        while subgraph_set != []:
            # determine next vertex set which will be added
            next_vertex_set = list(next(iterators[-1], []))
            # if now new vertex was created, create new iterator where vertex set of exactly one smaller size are searched, if the size gets to 0, return
            if next_vertex_set == []:
                if iterator_length[-1] > 1:
                    # in this case, create a new iterator of smaller size
                    iterator_length[-1] -= 1
                    iterators[-1] = combinations(neighbors_sets[-1], iterator_length[-1])
                    next_vertex_set = list(next(iterators[-1], []))
                    if next_vertex_set == []:
                        start = pointers.pop()
                        # delete all vertices added in this step to the subgraph set
                        max_vertex = -1
                        if pointers.__len__() > 0:
                            max_vertex = pointers[-1]
                        while start > max_vertex:
                            vertex_del = subgraph_set.pop()
                            current_graph.delete_vertices(sub_size_minus)
                            start -= 1
                            sub_size_minus -= 1
                            sub_set.pop(vertex_del)
                        neighbors_sets.pop()
                        iterator_length.pop()
                        iterators.pop()
                        continue
                else:
                    # otherwise no new vertex set can be added so return to the father
                    start = pointers.pop()
                    # delete all vertices added in this step to the subgraph set
                    max_vertex = -1
                    if pointers.__len__() > 0:
                        max_vertex = pointers[-1]
                    while start > max_vertex:
                        vertex_del = subgraph_set.pop()
                        current_graph.delete_vertices(sub_size_minus)
                        start -= 1
                        sub_size_minus -= 1
                        sub_set.pop(vertex_del)
                    neighbors_sets.pop()
                    iterator_length.pop()
                    iterators.pop()
                    continue
            # we found a new subgraph, so update all data structures
            nodes += 1
            # before doing this, check the time limit
            if clock() > time_max:
                return value, best_graph, nodes
            old_size = subgraph_set.__len__()
            subgraph_set.extend(next_vertex_set)
            current_graph.add_vertices(len(next_vertex_set))
            # determine all exclusive neighbors of the vertices in net_vertex_set
            neig_set = set()
            for vertex in next_vertex_set:
                sub_size_minus += 1
                sub_set.update({vertex: sub_size_minus})
            edges = []
            for vertex in next_vertex_set:
                for neig in graph.neighbors(vertex):
                    if neig > i:
                        break
                    neig_set.add(neig)
                    if neig in sub_set:
                        if neig in next_vertex_set:
                            if neig > vertex:
                                edges.append([sub_set[neig], sub_set[vertex]])
                        else:
                            edges.append([sub_set[neig], sub_set[vertex]])
            current_graph.add_edges(edges)
            # check if we found a solution
            if sub_size_minus == k_minus:
                new_value = ev.evaluate(current_graph)
                if new_value > value:
                    value = new_value
                    best_graph = current_graph.copy()
                    best_graph.vs["name"] = get_names(graph, subgraph_set)
                    # check if user_ub is matched
                    if ev.check_user_ub(value):
                        return value, best_graph, nodes
                # restore old subgraph set
                while sub_size_minus > pointers[-1]:
                    vertex_del = subgraph_set.pop()
                    current_graph.delete_vertices(sub_size_minus)
                    sub_size_minus -= 1
                    sub_set.pop(vertex_del)
            # otherwise check if adding this vertex can increase the value, if this is not possible, return to its parent
            elif ev.eval_add_vertices(graph.induced_subgraph(subgraph_set), k - old_size, get_names(graph, subgraph_set), max(graph.degree(next_vertex_set)), value):
                for i in range(next_vertex_set.__len__()):
                    vertex_del = subgraph_set.pop()
                    current_graph.delete_vertices(sub_size_minus)
                    sub_size_minus -= 1
                    sub_set.pop(vertex_del)
            # otherwise create the new child
            else:
                pointers.append(pointers[-1] + next_vertex_set.__len__())
                # delete all vertices which are currently in the subgraph set and all previous neighbors
                neig_set.difference_update(subgraph_set)
                for neigh_set in neighbors_sets:
                    neig_set.difference_update(neigh_set)
                # create new iterator, and new neighborset
                neighbors_sets.append(neig_set)
                iterator_length.append(min(k - subgraph_set.__len__(), neig_set.__len__()))
                iterators.append(combinations(neig_set, iterator_length[-1]))
        # now the root i and each vertex in that critical clique are forbidden
        update_forbidden_monotone(i,crit,children,graph,colors)
        twins.update_twin_relation_monotone(graph, crit, children, i,colors)
        del_edges = update_graph(graph, i, del_edges)
    return value, best_graph, nodes


# new version of kavosh without the break condition and iterative
def kavosh_iter_noreturn(graph, k, time_max, crit, value):
    # for each fixed vertex v find all connected subgraphs which include v
    # after the enumeration delete vertex v
    del_edges = []
    current_graph = Graph()
    best_graph = Graph()
    best_graph.vs["name"] = []
    nodes = 0
    k_minus = k - 1
    colors = graph.vcount() * [False]
    # consider only graphs with at least k vertices
    for i in range(graph.vcount()-1, k-2, -1):
        print_progress(i,graph.vcount(),value)
        if colors[i] == True:
            continue
        # save the subgraph set
        subgraph_set = [i]
        # subgraph set as set
        sub_set = {i: 0}
        sub_size_minus = 0
        current_graph.add_vertex()
        # set of pointers to save which enumeration tree node corresponds to which subgraph set
        pointers = [0]
        # neighbors sets
        neighs = []
        for neig in graph.neighbors(i):
            if neig > i:
                break
            neighs.append(neig)
        neighbors_sets = [neighs]
        # list of length of number of vertices which will be added if the next iterator is created
        iterator_length = [min(k-1, neighbors_sets[0].__len__())]
        # list of iterators
        iterators = [combinations(neighbors_sets[0], iterator_length[0])]
        # enumerate subgraphs while subgraph set is not empty
        while subgraph_set != []:
            # determine next vertex set which will be added
            next_vertex_set = list(next(iterators[-1], []))
            # if now new vertex was created, create new iterator where vertex set of exactly one smaller size are searched, if the size gets to 0, return
            if next_vertex_set == []:
                if iterator_length[-1] > 1:
                    # in this case, create a new iterator of smaller size
                    iterator_length[-1] -= 1
                    iterators[-1] = combinations(neighbors_sets[-1], iterator_length[-1])
                    next_vertex_set = list(next(iterators[-1], []))
                    if next_vertex_set == []:
                        start = pointers.pop()
                        # delete all vertices added in this step to the subgraph set
                        max_vertex = -1
                        if pointers.__len__() > 0:
                            max_vertex = pointers[-1]
                        while start > max_vertex:
                            vertex_del = subgraph_set.pop()
                            current_graph.delete_vertices(sub_size_minus)
                            start -= 1
                            sub_size_minus -= 1
                            sub_set.pop(vertex_del)
                        neighbors_sets.pop()
                        iterator_length.pop()
                        iterators.pop()
                        continue
                else:
                    # otherwise now ne vertex set can be added so return to the father
                    start = pointers.pop()
                    # delete all vertices added in this step to the subgraph set
                    max_vertex = -1
                    if pointers.__len__() > 0:
                        max_vertex = pointers[-1]
                    while start > max_vertex:
                        vertex_del = subgraph_set.pop()
                        current_graph.delete_vertices(sub_size_minus)
                        start -= 1
                        sub_size_minus -= 1
                        sub_set.pop(vertex_del)
                    neighbors_sets.pop()
                    iterator_length.pop()
                    iterators.pop()
                    continue
            # we found a new subgraph, so update all data structures
            nodes += 1
            # before doing this, check the time limit
            if clock() > time_max:
                return value, best_graph, nodes
            old_size = subgraph_set.__len__()
            subgraph_set.extend(next_vertex_set)
            current_graph.add_vertices(len(next_vertex_set))
            # determine all exclusive neighbors of the vertices in net_vertex_set
            neig_set = set()
            for vertex in next_vertex_set:
                sub_size_minus += 1
                sub_set.update({vertex: sub_size_minus})
            edges = []
            for vertex in next_vertex_set:
                for neig in graph.neighbors(vertex):
                    if neig > i:
                        break
                    neig_set.add(neig)
                    if neig in sub_set:
                        if neig in next_vertex_set:
                            if neig > vertex:
                                edges.append([sub_set[neig], sub_set[vertex]])
                        else:
                            edges.append([sub_set[neig], sub_set[vertex]])
            current_graph.add_edges(edges)
            # check if we found a solution
            if sub_size_minus == k_minus:
                new_value = ev.evaluate(current_graph)
                if new_value > value:
                    value = new_value
                    best_graph = current_graph.copy()
                    best_graph.vs["name"] = get_names(graph, subgraph_set)
                    # check if user_ub is matched
                    if ev.check_user_ub(value):
                        return value, best_graph, nodes
                # restore old subgraph set
                while sub_size_minus > pointers[-1]:
                    vertex_del = subgraph_set.pop()
                    current_graph.delete_vertices(sub_size_minus)
                    sub_size_minus -= 1
                    sub_set.pop(vertex_del)
            # otherwise check if adding this vertex can increase the value, if this is not possible, return to its parent
            elif ev.eval_add_vertices(graph.induced_subgraph(subgraph_set), k - old_size, get_names(graph, subgraph_set), max(graph.degree(next_vertex_set)), value):
                for i in range(next_vertex_set.__len__()):
                    vertex_del = subgraph_set.pop()
                    current_graph.delete_vertices(sub_size_minus)
                    sub_size_minus -= 1
                    sub_set.pop(vertex_del)
            # otherwise create the new child
            else:
                pointers.append(pointers[-1] + next_vertex_set.__len__())
                # delete all vertices which are currently in the subgraph set and all previous neighbors
                neig_set.difference_update(subgraph_set)
                for neigh_set in neighbors_sets:
                    neig_set.difference_update(neigh_set)
                # create new iterator, and new neighborset
                neighbors_sets.append(neig_set)
                iterator_length.append(min(k - subgraph_set.__len__(), neig_set.__len__()))
                iterators.append(combinations(neig_set, iterator_length[-1]))
        # now the root i and each vertex in that critical clique are forbidden
        update_forbidden(i,crit,graph, colors)
        twins.update_twin_relation(graph,crit,i, colors)
        del_edges = update_graph(graph, i, del_edges)
    return value, best_graph, nodes

# dictionary to choose the enumeration algorithm
enumeration_function = {'pivot-iter': pivot_iter_noreturn,
                       'pivot-iter_m': pivot_iter_noreturn_m,
                       'simple-iter': simple_iter_noreturn,
                       'simple-iter_m': simple_iter_noreturn_m,
                       'kavosh-iter': kavosh_iter_noreturn,
                       'kavosh-iter_m': kavosh_iter_noreturn_m,}

