from igraph import *
import random
import math
import src.eval_func as ev
import bisect
import sys
from time import *

# input is a list or a set of vertex IDs, output is a list of vertex names
def get_names(graph, vertex_ids):
    lis = []
    for id in vertex_ids:
        lis.append(graph.vs[id]["name"])
    return lis

def calc_random_lower_bound(graph, k, time_max):
    # all vertices of the subgraph set can have further neighbors, so hardcode still neighbors sets
    still_neig = [[0]]
    forb = [False]*graph.vcount()
    for i in range(1, k):
        still_neig.append(still_neig[-1][:])
        still_neig[-1].append(i)
    # adjacency lists as sets
    open_neigs = []
    for i in range(graph.vcount()):
        open_neigs.append(set(graph.neighbors(i)))
    # greedy heuristics to calculate good lower bounds for the solver
    # current best value
    value = -sys.maxsize + 10*k
    vertex_ub = ev.vertex_upper_bound
    best_graph = Graph()
    best_graph.vs["name"] = []
    iterations = int(math.log(graph.vcount(), 2)*k + 1)
    # initialize for the three distributions
    r_total = []
    r_sum_total = 0
    for i in range(graph.vcount()):
        r_sum_total += 1
        r_total.append(r_sum_total)
    m_total = []
    m_sum_total = 0
    for deg in graph.degree():
        m_sum_total += deg
        m_total.append(m_sum_total)

    maxdeg = graph.maxdegree() + 1
    d_total = []
    d_sum_total = 0
    for deg in graph.degree():
        d_sum_total += (maxdeg - deg)
        d_total.append(d_sum_total)

    values = [1]
    val_index = {1: 0}
    for i in range(1, k+1):
        new_entry = (i+1)**3
        values.append(new_entry)
        val_index.update({new_entry: i})
    for i in range(iterations):
        if clock() > time_max:
            return value, best_graph, False
        for h in ['r', 'm', 'd']:
            # pick one random start vertex
            if h == 'r':        
                s_vertex = weighted_choice_part(r_total, r_sum_total)
            elif h == 'm':
                s_vertex = weighted_choice_part(m_total, m_sum_total)
            else:
                s_vertex = weighted_choice_part(d_total, d_sum_total)
            vertex_set = {s_vertex: 0}
            sub_set = [s_vertex]
            size = 1
            # induced subgraph
            current_graph = Graph()
            current_graph.add_vertex()
            current_graph.vs[0]["id"] = s_vertex
            if h == 'r':
                # get neighbor set
                neighs_list = graph.neighbors(s_vertex)
                neighs_set = set(graph.neighbors(s_vertex))
                weights = []
                total_weight = 0
                for j in range(len(neighs_list)):
                    total_weight += 1
                    weights.append(total_weight)
            elif h == 'm':
                # get neighbor set
                neighs_list = []
                # dictionary for vertices and corresponding position in the neighs_list
                posi_in_list = dict()
                ind = 0
                weights = []
                org_weights = []
                total_weight = 0
                for neig in graph.neighbors(s_vertex):
                    neighs_list.append(neig)
                    posi_in_list.update({neig: ind})
                    ind += 1
                    total_weight += 1
                    weights.append(total_weight)
                    org_weights.append(1)
            else:
                # get neighbor set
                neighs_list = []
                # dictionary for vertices and corresponding position in the neighs_list
                posi_in_list = dict()
                ind = 0
                weights = []
                org_weights = []
                vertices_found = []
                total_weight = 0
                for neig in graph.neighbors(s_vertex):
                    neighs_list.append(neig)
                    posi_in_list.update({neig: ind})
                    ind += 1
                    total_weight += 1
                    weights.append(total_weight)
                    org_weights.append(1)
                    vertices_found.append(False)

            # now append vertex set to k vertices
            while size < k and len(weights) > size:
                if vertex_ub >= 0:
                    if ev.evaluate(current_graph, graph, forb, still_neig[size-1]) + (k-size)*vertex_ub <= value:
                        break
                # choose arbitrary vertex of the neighborhood
                vertex = neighs_list[weighted_choice_part(weights, total_weight)]
                while vertex in vertex_set:
                    vertex = neighs_list[weighted_choice_part(weights, total_weight)]
                vertex_set.update({vertex: size})
                sub_set.append(vertex)
                current_graph.add_vertex()
                current_graph.vs[size]["id"] = vertex
                edges = []
                # iterate over neighbors of new vertex
                for neig in graph.neighbors(vertex):
                    if h == 'r':
                        if neig not in neighs_set and neig is not s_vertex:
                            neighs_list.append(neig)
                            neighs_set.add(neig)
                            total_weight += 1
                            weights.append(total_weight)
                        if neig in vertex_set:
                            edges.append([vertex_set[neig], size])
                    elif h == 'm':
                        if neig not in posi_in_list and neig is not s_vertex:
                            neighs_list.append(neig)
                            posi_in_list.update({neig: ind})
                            total_weight += 1
                            weights.append(total_weight)
                            ind += 1
                            org_weights.append(1)
                        # if this neighbor is in the vertex set, adapt weights
                        if neig in vertex_set:
                            # get id and increase this value by one
                            start = vertex_set[neig]
                            index_val = val_index[org_weights[start]]
                            org_weights[start] = values[index_val+1]
                            edges.append([start, size])
                    else:
                        if neig not in posi_in_list and neig is not s_vertex:
                            neighs_list.append(neig)
                            posi_in_list.update({neig: ind})
                            total_weight += 1
                            weights.append(total_weight)
                            ind += 1
                            org_weights.append(1)
                            vertices_found.append(True)
                        # if this neighbor is in the vertex set, adapt weights
                        if neig in vertex_set:
                            # get id and increase this value by one
                            start = vertex_set[neig]
                            vertices_found[start] = True
                            edges.append([start, size])
            
                # update of all weights in one step
                if h =='m':
                    total_weight = 0
                    weights = []
                    for entry in org_weights:
                        total_weight += entry
                        weights.append(total_weight)
                # update of all weights in one step (only for false vertices)
                elif h == 'd':
                    for entry in range(len(vertices_found)):
                        if vertices_found[entry] is False:
                            org_weights[entry] = values[val_index[org_weights[entry]]+1]
                        else:
                            vertices_found[entry] = False
                    total_weight = 0
                    weights = []
                    for entry in org_weights:
                        total_weight += entry
                        weights.append(total_weight)
            
                        
                current_graph.add_edges(edges)
                size += 1
                # on size k evaluate
                if size == k:
                    new_value = ev.evaluate(current_graph, graph, forb, still_neig[-1])
                    if new_value > value:
                        value = new_value
                        best_graph = current_graph.copy()
                        best_graph.vs["name"] = get_names(graph, sub_set)
                        # now perform local search
                        best_graph, value = find_cutvertices(graph, best_graph, sub_set, value, time_max, open_neigs, still_neig[-1], forb)
                    # check if user_ub is matched
                    if ev.check_user_ub(value):
                        return value, best_graph, True
    return value, best_graph, False



# input is a list of weights, eg [10,2,3,5,80], then 0 is drawn with 0.1, 1 with 0.02, 2 with 0.03, 3 with 0.05 and 4 with 0.8
def weighted_choice_part(totals, running_total):
    rnd = random.random() * running_total
    return bisect.bisect_right(totals, rnd)

# local search for graph with better lower bound for the problem

# dfs-tree to find all cut points and all connected components after deleting such a cut-vertex (Tarjan-algo)
def DFS(graph, discover, low, vertex, dfs_counter, parent, children):
    discover[vertex] = dfs_counter
    low[vertex] = dfs_counter
    dfs_counter += 1
    for neig in graph.neighbors(vertex):
        if discover[neig] == -1:
            parent[neig] = vertex
            children[vertex].append(neig)
            discover, low, dfs_counter, parent, children = DFS(graph, discover, low, neig, dfs_counter, parent, children)
            low[vertex] = min(low[vertex], low[neig])
        elif neig != parent[vertex]:
            low[vertex] = min(low[vertex], discover[neig])
    return discover, low, dfs_counter, parent, children


def find_cutvertices(graph, curr_graph, sub_set, value, time_max, open_neigs, still_neig, forb):
    # first determine all biconnected components (without the cut-vertices) of curr_graph via DFS
    best_graph = Graph()
    best_graph.vs["name"] = []
    sub_set = set(sub_set)
    numb_vertices = len(sub_set)
    k_minus = numb_vertices - 1
    # found indicates which vertex was removed from the subgraph set and which one was added (if no improvement still -1)
    new_vertex = -1
    deleted_vertex = -1
    # do local search until no improvement can be made
    while True:
        discover = [-1] * numb_vertices
        lowest = [numb_vertices] * numb_vertices
        parents = [-1] * numb_vertices
        dfs_counter = 0
        children = [[] for i in range(numb_vertices)]
        for i in range(numb_vertices):
            if discover[i] == -1:
                discover, lowest, dfs_counter, parents, children = DFS(curr_graph, discover, lowest, i, dfs_counter, parents, children)
        low_vals = [[] for i in range(numb_vertices)]
        cut_list = [[] for i in range(numb_vertices)]
        for i in range(numb_vertices):
            low_vals[lowest[i]].append(i)
        # all neighbors of a vertex v with other low-number  will be cut off after the deletion of v
        for i in range(numb_vertices):
            for neig in curr_graph.neighbors(i):
                if lowest[neig] != lowest[i]:
                    cut_list[i].append(neig)
        # now check each subgraph T:=S-{u}\cup{v} such that T is connected
        # check if any such T has a better objective value than S
        # root or vertex=0 needs another consideration, there the start differs
        # if root 0 has only 1 child, root is no cut-vertex, otherwise different start
        start = 1
        if len(children[0]) == 1:
            start = 0
        if start == 1:
            neigs = set()
            done_lowest = [False] * numb_vertices
            done_lowest[0] = True
            flag = False
            for childd in children[0]:
                child_stack = [childd]
                new_neigs = set()
                rem_compo = []
                while child_stack != []:
                    vertexx = child_stack.pop()
                    new_neigs.update(graph.neighbors(graph.vs.find(name=curr_graph.vs[vertexx]["name"])))
                    for neighbo in children[vertexx]:
                        neig_low = lowest[neighbo]
                        if neighbo != 0:
                            if neig_low == 0:
                                child_stack.append(neighbo)
                            else:
                                rem_compo.append(neig_low)
                                done_lowest[neig_low] = True
                # initial subset of vertices with low=0 is now computed
                # further as below
                while rem_compo != []:
                    low_compo = rem_compo.pop()
                    for vertex in low_vals[low_compo]:
                        # increase neighborhood of that component
                        new_neigs.update(graph.neighbors(graph.vs.find(name=curr_graph.vs[vertex]["name"])))
                        # get children of that vertex to other low-values and add them to rem_compo
                        for child in cut_list[vertex]:
                            if done_lowest[lowest[child]] == False:
                                rem_compo.append(lowest[child])
                                done_lowest[lowest[child]] = True
                if flag == True:
                    neigs.intersection_update(new_neigs)
                    # break if this set is empty
                    if neigs == set():
                        break
                else:
                    neigs = new_neigs
                    flag = True
            # if flag is False, then the remaining component was empty
            # in that case only calculate common neighborhood of the children of i
            # otherwise calculate also common neighborhood with neigs
            for child in cut_list[i]:
                if done_lowest[lowest[child]] == True:
                    continue
                new_neigs = set()
                rem_compo = [lowest[child]]
                done_lowest[lowest[child]] = True
                while rem_compo != []:
                    low_compo = rem_compo.pop()
                    for vertex in low_vals[low_compo]:
                        # increase neighborhood of that component
                        new_neigs.update(graph.neighbors(graph.vs.find(name=curr_graph.vs[vertex]["name"])))
                        # get children of that vertex to other low-values and add them to rem_compo
                        for child in cut_list[vertex]:
                            if done_lowest[lowest[child]] == False:
                                rem_compo.append(lowest[child])
                                done_lowest[lowest[child]] = True
                neigs.intersection_update(new_neigs)
            # now remove all vertices of curr_subgraph
            neigs.difference_update(sub_set)
            # now create a copy of the current subgraph, and remove vertex i from that copy
            # for each vertex in neigs, add this vertex to the copy and evaluate this subgraph
            test_graph = curr_graph.copy()
            test_graph.delete_vertices(0)
            name_i = curr_graph.vs[0]["name"]
            # now for each common neighbor create a new subgraph and test it
            test_graph.add_vertex()
            for vertex in neigs:
                test_graph.vs[k_minus]["name"] = graph.vs[vertex]["name"]
                test_graph.vs["id"] = vertex
                # add edges
                edges = []
                vertices_with_edges = open_neigs[vertex].intersection(sub_set)
                for vertexx in vertices_with_edges:
                    name_vertexx = graph.vs[vertexx]["name"]
                    if name_vertexx != name_i:
                        edges += [(k_minus, test_graph.vs.find(name=name_vertexx))]
                test_graph.add_edges(edges)
                # now evaluate and update if better objective value
                new_val = ev.evaluate(test_graph, graph, forb, still_neig)
                if value < new_val:
                    value = new_val
                    best_graph = test_graph.copy()
                    # update new_vertex and deleted_vertex
                    deleted_vertex = graph.vs.find(name=curr_graph.vs[0]["name"]).index
                    new_vertex = vertex
                # delete last vertex
                test_graph.delete_edges(edges)
        for i in range(start, numb_vertices):
            # calculate initial neighbor set of remaining low-component (consider empty-case separately)
            neigs = set()
            flag = False
            done_lowest = [False]*numb_vertices
            # calculate neighborhood of the component of vertex i (after its deletion its component is the set of vertices with same low-value and there subsequent children)
            rem_compo = [lowest[i]]
            done_lowest[lowest[i]] = True
            while rem_compo != []:
                low_compo = rem_compo.pop()
                for vertex in low_vals[low_compo]:
                    if vertex != i:
                        flag = True
                        # increase neighborhood of that component
                        neigs.update(graph.neighbors(graph.vs.find(name=curr_graph.vs[vertex]["name"])))
                        # get children of that vertex to other low-values and add them to rem_compo
                        for child in cut_list[vertex]:
                            if done_lowest[lowest[child]] == False:
                                rem_compo.append(lowest[child])
                                done_lowest[lowest[child]] = True
            # if flag is False, then the remaining component was empty
            # in that case only calculate common neighborhood of the children of i
            # otherwise calculate also common neighborhood with neigs
            for child in cut_list[i]:
                if done_lowest[lowest[child]] == True:
                    continue
                new_neigs = set()
                rem_compo = [lowest[child]]
                done_lowest[lowest[child]] = True
                while rem_compo != []:
                    low_compo = rem_compo.pop()
                    for vertex in low_vals[low_compo]:
                        # increase neighborhood of that component
                        new_neigs.update(graph.neighbors(graph.vs.find(name=curr_graph.vs[vertex]["name"])))
                        # get children of that vertex to other low-values and add them to rem_compo
                        for child in cut_list[vertex]:
                            if done_lowest[lowest[child]] == False:
                                rem_compo.append(lowest[child])
                                done_lowest[lowest[child]] = True
                if flag == True:
                    neigs.intersection_update(new_neigs)
                    # break if this set is empty
                    if neigs == set():
                        break
                else:
                    neigs = new_neigs
                    flag = True
            # now remove all vertices of curr_subgraph
            neigs.difference_update(sub_set)
            # now create a copy of the current subgraph, and remove vertex i from that copy
            # for each vertex in neigs, add this vertex to the copy and evaluate this subgraph
            test_graph = curr_graph.copy()
            test_graph.delete_vertices(i)
            name_i = curr_graph.vs[i]["name"]
            # now for each common neighbor create a new subgraph and test it
            test_graph.add_vertex()
            for vertex in neigs:
                test_graph.vs[k_minus]["name"] = graph.vs[vertex]["name"]
                test_graph.vs["id"] = vertex
                # add edges
                edges = []
                vertices_with_edges = open_neigs[vertex].intersection(sub_set)
                for vertexx in vertices_with_edges:
                    name_vertexx = graph.vs[vertexx]["name"]
                    if name_vertexx != name_i:
                        edges += [(k_minus, test_graph.vs.find(name=name_vertexx))]
                test_graph.add_edges(edges)
                # now evaluate and update if better objective value
                new_val = ev.evaluate(test_graph, graph, forb, still_neig)
                if value < new_val:
                    value = new_val
                    best_graph = test_graph.copy()
                    # update new_vertex and deleted_vertex
                    deleted_vertex = graph.vs.find(name=curr_graph.vs[i]["name"]).index
                    new_vertex = vertex
                # delete last vertex
                test_graph.delete_edges(edges)
        # break if value is not higher than before (or equivalently new_vertes = -1)
        if new_vertex == -1 or clock() > time_max:
            return curr_graph, value
        # a better solution was found, so reset and update curr_graph
        else:
            sub_set.remove(deleted_vertex)
            sub_set.add(new_vertex)
            curr_graph = best_graph.copy()
            new_vertex = -1
            deleted_vertex = -1
