from __future__ import print_function
from igraph import *
from time import *
import sys


# reads a graph and returns an array with the critical cliques, only if k>1

def crit_cliques_edge_monotone(graph, k, time_max):
    # vertices which will be deleted from the graph
    deleted = set()
    # original indices
    n = graph.vcount()
    n3 = n**3
    indices = range(n)
    indices2 = range(n)
    deg = graph.degree()
    # array with neighborhoods
    neighborhoods =  dict()
    open_neighborhoods = dict()
    # value and values will be our hashes for closed and open neighborhood
    # let v\in V: value(v)=|N[v]| + |N[N[v]]|*n + sum_v in N[v](id(v))
    values = []
    values2 = []
    value_dict = dict()
    for index in indices:
        val = deg[index]
        degs=0
        ids=0
        for vertex in graph.neighbors(index):
            degs += deg[vertex]
            ids += vertex
        val += n*degs +ids*n3
        values.append(val +index*n3)
        values2.append(val)
        value_dict[index]=val
        neighborhoods[index] = (set(graph.neighbors(index)).union({index}))
        open_neighborhoods[index] = (set(graph.neighbors(index)))
    # sort the vertices according to their hash value
    sizes = sorted(list(zip(values, indices)), reverse=True)
    values, indices = list(zip(*sizes))
    
    # If two vertices are in the same critical clique, they have the same value
    # in the following check for two subsequent vertices with the same value if they in fact have the same closed neighborhood
    index = 0
    neighbors = [set(graph.neighbors(indices[0])).union({indices[0]})]
    value = values[0]
    # the maximum size of a critical clique is k, all remaining vertices in this critical clique can be removed
    size = [1]
    unset_vertices = [indices[0]]
    used = set()
    numb_crit_cliques = 0
    crit_indices = [0]
    for i in range(1, n):
        val_i = values[i]
        vertex = indices[i]
        if value == val_i:
            # determine neighborhood of this vertex
            new_neigs = neighborhoods[vertex]
            # test if this new neighborhood was already observed
            flag = False
            lint = 0
            for neig in neighbors:
                if neig == new_neigs:
                    if size[lint] >= k:
                        deleted.add(vertex)
                    else:
                        if size[lint] == 1:
                            crit_indices[lint] = numb_crit_cliques
                            graph.vs[unset_vertices[lint]]["crit"] = index + numb_crit_cliques
                            used.add(unset_vertices[lint])
                            numb_crit_cliques += 1
                        graph.vs[vertex]["crit"] = index + crit_indices[lint]
                        used.add(vertex)
                    size[lint] += 1
                    flag = True
                    break
                lint += 1
            if flag == False:
                neighbors.append(new_neigs)
                size.append(1)
                unset_vertices.append(vertex)
                crit_indices.append(0)
        else:
            neighbors = [set(graph.neighbors(vertex)).union({vertex})]
            value = val_i
            index += numb_crit_cliques
            numb_crit_cliques = 0
            size = [1]
            unset_vertices = [vertex]
            crit_indices = [0]

    sizes = sorted(zip(values2, indices2), reverse=True)
    values2, indices = list(zip(*sizes))
    # find first vertex that is not a critical clique
    i = 0
    while indices[i] in deleted or indices[i] in used:
        i += 1
    index += numb_crit_cliques
    neighbors = [set(graph.neighbors(indices[i]))]
    value = values2[i]
    graph.vs[indices[i]]["crit"] = index
    size = [1]
    k -= 1
    for i in range(i+1, n):
        vertex = indices[i]
        if vertex in deleted or vertex in used:
            continue
        val_i = values2[i]
        if value == val_i:
            # determine neighborhood of this vertex
            new_neigs = open_neighborhoods[vertex]
            flag = False
            lint = 0
            for neig in neighbors:
                if neig == new_neigs:
                    flag = True
                    if size[lint] >= k:
                        deleted.add(vertex)
                    else:
                        graph.vs[vertex]["crit"] = index + lint
                        used.add(vertex)
                        size[lint] += 1
                    break
                lint += 1
            if flag == False:
                neighbors.append(new_neigs)
                size.append(1)
                graph.vs[vertex]["crit"] = index + lint
        else:
            neighbors = [open_neighborhoods[vertex]]
            value = val_i
            index += size.__len__()
            size = [1]
            graph.vs[vertex]["crit"] = index

    # delete vertices such that each critical clique has size at most k and each IS size at most k-1
    # update: deletion is now done at the very end
    # graph.delete_vertices(deleted)

    indices = range(n)
    new_deleted = set(deleted)
    
    deg = graph.degree()
    sizes = sorted(list(zip(deg, indices)), reverse=True)
    unzipped = list(zip(*sizes))
    indices = unzipped[1]
    # now create an array which contains an array where each entry represents a critical clique/IS
    # this entry contains all vertices in this critical clique/IS
    index += size.__len__()
    cliques = [[] for i in range(index)]
    i = 0
    for vertex in graph.vs:
        if vertex["crit"] == None:
            continue
        cliques[vertex["crit"]].append(i)
        i += 1
    # data structure for testing subsets of neighborhoods,
    # for vertex u, list u contains all vertices v with N(v)\subset N(u)
    n_range = range(n)
    children = [[] for i in n_range]
    structure = [[] for i in n_range]
    visited = [False] * n
    # initially roots can be empty since indices is an ordering of decreasing degree
    roots = []
    root_set = set()
    print("Computed twin sets in",clock(),"seconds.")
    
    for i in n_range:
        # check running time after each vertex
        if clock() > time_max:
            return graph, cliques, structure, False
        vertex = indices[i]
        if vertex in new_deleted:
            continue
        new = []
        if True:
            if graph.degree(vertex)!=0:                
                x = graph.neighbors(vertex)[0]
                for u in graph.neighbors(vertex):
                    if graph.degree(u)<graph.degree(x):
                        x = u
                root_cand = set(neighborhoods[x])
                root_cand.remove(vertex)
                root_cand = root_cand.intersection(root_set)
                for u in graph.neighbors(vertex):
                    root_cand = root_cand.intersection(neighborhoods[u])
                    if len(root_cand) == 0:
                        break                
                for r in root_cand:
                    new.append(r)
        else:
            for ii in range(roots.__len__()):
                # necessary conditions: vertex has smaller degree and a smaller hash
                if graph.degree(vertex) < graph.degree(roots[ii]) and value_dict[vertex] <= value_dict[roots[ii]] :
                    if open_neighborhoods[vertex] < neighborhoods[roots[ii]]:
                        new.append(roots[ii])
        # if new is still empty, create a new root
        if new == []:
            root_set.add(vertex)
            # do not do all these checks again for each vertex in the same critical clique
            for ver in cliques[graph.vs[vertex]["crit"]]:
                new_deleted.add(ver)
        # else determine all vertices which are supersets of vertex
        else:
            uused = []
            while new != []:
                this_node = new.pop()
                # check all direct children of this_node, if there are unvisited vertices and check if neighborhood of vertex is still a subset
                # if yes, add these vertices to the stack and if not add this vertex to the children array
                flag = False
                for sub_vertex in children[this_node]:
                    # necessary condition: smaller degree and not larger hash
                    if graph.degree(vertex) < graph.degree(sub_vertex) and value_dict[vertex] <= value_dict[sub_vertex]:
                        if open_neighborhoods[vertex] < neighborhoods[sub_vertex]:
                            if visited[sub_vertex] == False:
                                new.append(sub_vertex)
                            visited[sub_vertex] = True
                            uused.append(sub_vertex)
                            flag = True
                if flag == False:
                    children[this_node].append(vertex)
                # add structure for all vertices in that critical clique
                # append for every vertex in the critical clique of this_node
                for crit_node in cliques[graph.vs[this_node]["crit"]]:
                    for ver in cliques[graph.vs[vertex]["crit"]]:
                        structure[crit_node].append(ver)
            # now forbid all vertices of that critical clique
            for ver in cliques[graph.vs[vertex]["crit"]]:
                new_deleted.add(ver)
            # undo uused
            for entry in uused:
                visited[entry] = False

    graph.delete_vertices(deleted)

    print("Computed domination DAG in",clock(),"seconds.")
    return graph, cliques, structure, True



def crit_cliques(graph, k):
    # vertices which will be deleted from the graph
    deleted = set()
    # original indices
    n = graph.vcount()
    n3 = n**3
    indices = range(n)
    indices2 = range(n)
    deg = graph.degree()
    # value and values will be our hashes for closed and open neighborhood 
    # let v\in V: value(v)=|N[v]| + |N[N[v]]|./(|V|^2)
    values = []
    values2 = []
    for index in indices:
        val = deg[index]
        degs = 0
        ids = 0
        for vertex in graph.neighbors(index):
            degs += deg[vertex]
            ids += vertex
        val += degs*n + ids*n3
        values.append(val+ index*n3)
        values2.append(val)

    # sort the vertices according to their hash value
    sizes = sorted(list(zip(values, indices)), reverse=True)
    values, indices = list(zip(*sizes))
    
    # If two vertices are in the same critical clique, they have the same value
    # in the following check for two subsequent vertices with the same value if they in fact have the same closed neighborhood
    index = 0
    neighbors = [set(graph.neighbors(indices[0])).union({indices[0]})]
    value = values[0]
    # the maximum size of a critical clique is k, all remaining vertices in this critical clique can be removed
    size = [1]
    unset_vertices = [indices[0]]
    used = set()
    numb_crit_cliques = 0
    crit_indices = [0]
    for i in range(1, n):
        val_i = values[i]
        vertex = indices[i]
        if value == val_i:
            # determine neighborhood of this vertex
            new_neigs = set(graph.neighbors(vertex)).union({vertex})
            # test if this new neighborhood was already observed
            flag = False
            lint = 0
            for neig in neighbors:
                if neig == new_neigs:
                    if size[lint] >= k:
                        deleted.add(vertex)
                    else:
                        if size[lint] == 1:
                            crit_indices[lint] = numb_crit_cliques
                            graph.vs[unset_vertices[lint]]["crit"] = index + numb_crit_cliques
                            used.add(unset_vertices[lint])
                            numb_crit_cliques += 1
                        graph.vs[vertex]["crit"] = index + crit_indices[lint]
                        used.add(vertex)
                    size[lint] += 1
                    flag = True
                    break
                lint += 1
            if flag == False:
                neighbors.append(new_neigs)
                size.append(1)
                unset_vertices.append(vertex)
                crit_indices.append(0)
        else:
            neighbors = [set(graph.neighbors(vertex)).union({vertex})]
            value = val_i
            index += numb_crit_cliques
            numb_crit_cliques = 0
            size = [1]
            unset_vertices = [vertex]
            crit_indices = [0]

    sizes = sorted(zip(values2, indices2), reverse=True)
    values2, indices = zip(*sizes)

    # find first vertex that is not a critical clique
    i = 0
    while indices[i] in deleted or indices[i] in used:
        i += 1
    index += numb_crit_cliques
    neighbors = [set(graph.neighbors(indices[i]))]
    value = values2[i]
    graph.vs[indices[i]]["crit"] = index
    size = [1]
    k -= 1
    for i in range(i+1, n):
        vertex = indices[i]
        if vertex in deleted or vertex in used:
            continue
        val_i = values2[i]
        if value == val_i:
            # determine neighborhood of this vertex
            new_neigs = set(graph.neighbors(vertex))
            flag = False
            lint = 0
            for neig in neighbors:
                if neig == new_neigs:
                    flag = True
                    if size[lint] >= k:
                        deleted.add(vertex)
                    else:
                        graph.vs[vertex]["crit"] = index + lint
                        used.add(vertex)
                        size[lint] += 1
                    break
                lint += 1
            if flag == False:
                neighbors.append(new_neigs)
                size.append(1)
                graph.vs[vertex]["crit"] = index + lint
        else:
            neighbors = [set(graph.neighbors(vertex))]
            value = val_i
            index += size.__len__()
            size = [1]
            graph.vs[vertex]["crit"] = index

    print("Computed twin sets in", clock(), "seconds.")
    # delete vertices such that each critical clique has size at most k and each IS size at most k-1
    graph.delete_vertices(deleted)
    index += size.__len__()
    # now create an array which contains an array where each entry represents a critical clique/IS
    # this entry contains all vertices in this critical clique/IS
    cliques = [[] for i in range(index)]
    i = 0
    for vertex in graph.vs:
        cliques[vertex["crit"]].append(i)
        i += 1
    return graph, cliques


# update the twin relation after some vertex v has been discarded in the main enumeration loop
def update_twin_relation_monotone(graph, crit, children, v, forb_vertices):
    # auxiliary set for storing all vertices in N(v) whose twins have been processed
    touched = set()
    neighs = dict()
    closed_neighs = dict()
    neighs[v] = set(graph.neighbors(v))
    # interesting only if one neighbor and one further vertex in second/third neighborhood not yet forbidden
    # if two twin classes may be merged, then exactly one of them contains a neighbor u of v
    for u in graph.neighbors(v):
        neighs[u] = set(graph.neighbors(u))
        remove_forbidden(neighs[u],forb_vertices)
        closed_neighs[u] = set(neighs[u]).union({u})
        # the other one contains a neighbor of u or a neighbor of a neighbor of u
        cand = set()
        min_w = 0
        w_deg = sys.maxsize
        crit_u = graph.vs[u]["crit"]
        closed_u = closed_neighs[u]
        for w in graph.neighbors(u):
            deg_w=graph.degree(w)
            if deg_w >= len(neighs[u]) and w not in touched and w not in neighs[v] and crit_u !=graph.vs[w]["crit"]:
                cand.add(w)
            if deg_w < w_deg and not forb_vertices[w]:
                min_w = w
                w_deg = deg_w
        for x in graph.neighbors(min_w):
            if x!=u and graph.degree(x) >= len(neighs[u]) and x not in touched and crit_u !=graph.vs[x]["crit"] and x not in neighs[v] and x not in neighs[u]:
                cand.add(x)
        for w in cand:
            if w not in neighs:
                neighs[w] = set(graph.neighbors(w))
                remove_forbidden(neighs[w],forb_vertices)
                if len(neighs[w])!=len(neighs[u]):
                    {} # continue
            if w in neighs[u] and w not in closed_neighs:
                closed_neighs[w] = set(neighs[w]).union({w})
            twin_compatible = False
            domination_compatible = False
            if w in neighs[u] and closed_neighs[w]==closed_u: twin_compatible = True 
            elif neighs[w] == neighs[u]: twin_compatible = True
            if twin_compatible:
                touched.add(w)
                touched.add(u)
                # update forbidden ds if exactly one of them is forbidden
                lw = crit[graph.vs[w]["crit"]]
                lu = crit[crit_u]
                if forb_vertices[u] and not forb_vertices[w]:
                    for x in lw:
                        forb_vertices[x]=True
                elif forb_vertices[w] and not forb_vertices[u]:
                    for x in lu:
                        forb_vertices[x]=True
                # merge twin sets
                if len(lu)>len(lw):
                    lu.extend(lw)
                    for x in lw:
                        graph.vs[x]["crit"]=graph.vs[u]["crit"]
                else:
                    lw.extend(lu)
                    for x in lu:
                        graph.vs[x]["crit"]=graph.vs[w]["crit"]
                break
            if u not in children[w] and len(neighs[w])>len(neighs[u]):
                if w in neighs[u] and closed_neighs[w]>=closed_u:
                    domination_compatible = True
                if w not in neighs[u] and neighs[w]>=neighs[u]:
                    domination_compatible = True
            if domination_compatible:
                {}
                children[w].append(u)
        touched.add(u)

# update the twin relation after some vertex v has been discarded in the main enumeration loop
def update_twin_relation(graph, crit, v, forb_vertices):
    # auxiliary set for storing all vertices in N(v) whose twins have been processed
    touched = set()
    neighs = dict()
    closed_neighs = dict()
    neighs[v] = set(graph.neighbors(v))
    # interesting only if one neighbor and one further vertex in second/third neighborhood not yet forbidden
    # if two twin classes may be merged, then exactly one of them contains a neighbor u of v
    for u in graph.neighbors(v):
        # if u<v: continue
        neighs[u] = set(graph.neighbors(u))
        remove_forbidden(neighs[u],forb_vertices)
        closed_neighs[u] = set(neighs[u]).union({u})
        # the other one contains a neighbor of u or a neighbor of a neighbor of u
        cand = set()
        min_w = 0
        w_deg = sys.maxsize
        crit_u = graph.vs[u]["crit"]
        closed_u = closed_neighs[u]
        for w in graph.neighbors(u):
            deg_w = graph.degree(w)
            if deg_w >= len(neighs[u]) and w not in touched and w not in neighs[v] and crit_u !=graph.vs[w]["crit"]:
                cand.add(w)
            if deg_w < w_deg and not forb_vertices[w]:
                min_w = w
                w_deg = deg_w
        for x in graph.neighbors(min_w):
            if x != u and graph.degree(x)>=len(neighs[u]) and x not in touched and crit_u !=graph.vs[x]["crit"] and x not in neighs[v] and x not in neighs[u]:
                cand.add(x)
        for w in cand:
            if w not in neighs:
                neighs[w] = set(graph.neighbors(w))
                remove_forbidden(neighs[w],forb_vertices)
                if len(neighs[w]) != len(neighs[u]):
                    continue
            if w in neighs[u] and w not in closed_neighs:
                closed_neighs[w] = set(neighs[w]).union({w})
            compatible = False
            if w in neighs[u] and closed_neighs[w]==closed_u: compatible = True 
            elif neighs[w] == neighs[u]: compatible = True
            if compatible:
                touched.add(w)
                touched.add(u)
                # update forbidden ds if exactly one of them is forbidden
                lw = crit[graph.vs[w]["crit"]]
                lu = crit[crit_u]
                if forb_vertices[u] and not forb_vertices[w]:
                    for x in lw:
                        forb_vertices[x] = True
                elif forb_vertices[w] and not forb_vertices[u]:
                    for x in lu:
                        forb_vertices[x] = True
                # merge twin sets
                if len(lu)>len(lw):
                    lu.extend(lw)
                    for x in lw:
                        graph.vs[x]["crit"]=graph.vs[u]["crit"]
                else:
                    lw.extend(lu)
                    for x in lu:
                        graph.vs[x]["crit"]=graph.vs[w]["crit"]
                break
        touched.add(u)

    

def remove_forbidden(vertex_set,forb_vertices):
    diff = set()
    for v in vertex_set:
        if forb_vertices[v]:
            diff.add(v)
    vertex_set.difference_update(diff)
